
from baopig import *


# ------ TESTS TO ENFORCE ------
# TEST 01 : Any component from a GridLayer require 'row' and 'col' attributes
# TEST 02 : Can't define 'pos' and 'origin' attributes of components who will be stored in a grid
# TEST 03 : Default behavior is to create rows and columns automatically
# TEST 04 : When the nbrows is set, we can't add a component who would like to go outside, same for nbcols
# TEST 05 : A row without defined height adapt to its components, 0 if empty, same for columns
# TEST 06 : We can set a default size for rows and columns
# TEST 07 : A component's hitbox is always inside its cell -> the cell defines the window
# TEST 08 : We can resize a row without any visual bug inside the row, same for columns -> the window is updated
# TEST 09 : Resizing a row moves the rows located below, same for columns
# TEST 10 : Components in a grid can't manage their position themself (non-dragable)

# ------ TESTS TO APPLY ------
# TODO : write unit tests with the removable rects

# TODO : rethink : can a component located in a grid move itself (via Dragable for example)

class UT_GridLayer_Scene(Zone):

    def __init__(self, *args, **kwargs):
        Zone.__init__(self, *args, **kwargs)

        self.set_mode(RESIZABLE)

        Layer(self, "zones_layer", adaptable=False)  # TODO : test adaptable
        z1 = Zone(self, size=(self.w-20, 130), background_color=(150, 150, 150), pos=(10, 10))
        z2 = Zone(self, size=(self.w-20, 130), background_color=(150, 150, 150),
                  origin=Origin(pos=(0, 10), reference_comp=z1, reference_location="bottomleft"))
        z3 = Zone(self, size=(self.w-20, 400), background_color=(150, 150, 150),
                  origin=Origin(pos=(0, 10), reference_comp=z2, reference_location="bottomleft"))

        z1.grid = GridLayer(z1, "grid_layer", row_height=40, col_width=100)
        assert z1.layers_manager.default_layer == z1.grid
        for row in range(2):
            for col in range(5):
                text = "row:{}\ncol:{}".format(row, col)
                Text(z1, text, row=row, col=col, name=text)

        z2.grid = GridLayer(z2, "grid_layer")
        for row in range(3):
            for col in range(5):
                text = "row:{}\ncol:{}".format(row, col)
                Text(z2, text, row=row, col=col, name=text)
        r = Rectangle(z2, (130, 49, 128), size=(30, 30), col=5, row=0)
        Dragable.set_dragable(r)
        Text(z2, "HI", col=6, row=0)
        Text(z2, "HI", col=7, row=1)
        r = Rectangle(z2, (130, 49, 128), size=(30, 30), col=8, row=2)
        Dragable.set_dragable(r)
        Button(z2, "Update sizes", command=z2.grid._update_size, col=0, row=3)

        grid = GridLayer(z3, "grid_layer", nbrows=3, nbcols=3)
        class RemovableRect(Rectangle, Linkable):
            def __init__(self, *args, **kwargs):
                Rectangle.__init__(self, *args, **kwargs)
                Linkable.__init__(self)
                def update(*args):
                    if self.collidemouse() and isinstance(mouse.linked_comp, RemovableRect):
                        self.kill()
                self.connect("kill", self.signal.LINK)
                mouse.signal.DRAG.add_command(update)
                self.parent.signal.RESIZE.add_command(update)
        import random
        color = lambda: [int(random.random() * 255)] * 2 + [128]
        def toggle_col_size(col_index):
            col = grid.get_col(col_index)
            if col.is_adaptable:  col.set_width(40)
            elif col.width == 40: col.set_width(20)
            else:                 col.set_width(None)
            if col.is_adaptable:  col[-1].kill()
        def toggle_row_size(row_index):
            row = grid.get_row(row_index)
            if row.is_adaptable:   row.set_height(40)
            elif row.height == 40: row.set_height(20)
            else:                  row.set_height(None)
            if row.is_adaptable:   row[-1].kill()
        def add_rect():
            for row in range(grid.nbrows):
                for col in range(grid.nbcols):
                    if grid._data[row][col] is None:
                        if row is grid.nbrows-1:
                            Button(z3, "TOG", row=row, col=col,
                                   command=PrefilledFunction(toggle_col_size, col),
                                   catch_errors=False, w=30, h=30)
                        elif col is grid.nbcols-1:
                            Button(z3, "TO\nG", row=row, col=col,
                                   command=PrefilledFunction(toggle_row_size, row),
                                   catch_errors=False, w=30, h=30)
                        else:
                            RemovableRect(z3, color(), (30, 30), col=col, row=row)
        Button(z3, "ADD", row=0, col=0, command=add_rect, catch_errors=True, w=30, h=30)
        def fix():
            if grid.cols_are_adaptable:
                grid.set_row_height(30)
                grid.set_col_width(30)
            else:
                grid.set_row_height(None)
                grid.set_col_width(None)
        Button(z3, "FIX", row=0, col=1, command=fix, catch_errors=False)

        # self.handle_event[mouse.LEFTBUTTON_DOWN].add(print)
        # self.handle_event[mouse.DRAG].add(print)
        # self.handle_event[mouse.RELEASEDRAG].add(print)


ut_scenes = [
    UT_GridLayer_Scene,
]


if __name__ == "__main__":
    from baopig.unit_tests.TesterScene import TesterScene
    app = Application()
    for scene in ut_scenes:
        TesterScene(app, scene)
    app.launch()







import pygame
from baopig._lib import Clickable, Hoverable
from baopig._lib import Rectangle
from .button import Button, ButtonText


class CheckMark(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        height=7,
        width=7,
        color="theme-color-border"
    )

    def __init__(self, checkbox):

        # mid = checkbox.height / 2
        Rectangle.__init__(self, checkbox, pos=(0, 0), pos_location="center",
                           pos_ref=checkbox.checkmarkframe, pos_ref_location="center")


class CheckMarkFrame(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        height=15,
        width=15,
        border_width=2,
        color=(0, 0, 0, 0)
    )

    def __init__(self, checkbox):

        mid = checkbox.height / 2
        Rectangle.__init__(self, checkbox, pos=(checkbox.padding.left, mid), pos_location="midleft")


class CheckBox(Button):
    # NOTE : connect a function to any state change with the VALIDATE signal

    STYLE = Button.STYLE.substyle()
    STYLE.modify(
        width=100,
        height=35,
        background_color = (0, 0, 0, 0),
        text_style = {"font_height": 15},
        padding=5
    )
    # TODO : rework checkbox's margin padding and this kind of stuff
    STYLE.create(
        checkmark_class = CheckMark,
        checkmarkframe_class = CheckMarkFrame
    )

    def __init__(self, parent, text=None, is_selected=False, **kwargs):

        Button.__init__(
            self, parent, text=text, **kwargs
        )

        self._is_selected = is_selected

        self._checkmarkframe_ref = CheckMarkFrame(self).get_weakref()
        self._checkmark_ref = CheckMark(self).get_weakref()

        self.checkmark.hide()

        # TODO : all of this, rework with better padding and margin management
        # TODO : when the text is too long, and the width is not specified, the box is lengthen

        # Sliding the text to the right
        assert self.w > self.h
        self.text_widget.center = ((self.w + self.checkmarkframe.right) / 2, self.h / 2)
        area = pygame.Rect(self.content_rect)
        area_end = area.right
        area.left = self.checkmarkframe.right + self.checkmarkframe.left
        area.w = area_end - area.left
        while self.text_widget.w > area.w:
            if self.text_widget.font.height == 2:
                raise ValueError(f"This text is too long for the text area : {text} (area={rect})")
            self.text_widget.font.config(height=self.text_widget.font.height - 1)

    checkmark = property(lambda self: self._checkmark_ref())
    checkmarkframe = property(lambda self: self._checkmarkframe_ref())
    is_selected = property(lambda self: self._is_selected)

    def validate(self, *args, **kwargs):

        self._is_selected = not self.is_selected

        if self.is_selected is True:
            self.checkmark.show()
        else:
            self.checkmark.hide()

        self.command(*args, **kwargs)

        return self.is_selected





import threading
import pygame
from baopig.pybao.issomething import *
from baopig.pybao.objectutilities import *
from baopig._debug import debug_screen_updates
from baopig.io import LOGGER, mouse, keyboard
from .utilities import *
from .widget import Widget
from .resizable import ResizableWidget
from .layersmanager import LayersManager


class Container(ResizableWidget):  # TODO : philosophy : is it good to force all containers to have a resize() method ?
                                   # TODO : where is it usefull and where is it a problem ?
                                   # A problem for Button who should never be resized ? right ?
    """
    Abstract class for widgets who need to contain other widgets

    We need the self.container_[action]() functions for recursivity between Container,
    because a container can contain an Openable without being a Openable himself

    WARNING : Try to do not override 'container_something' methods
    """

    STYLE = ResizableWidget.STYLE.substyle()
    STYLE.create(
        background_color = (0, 0, 0, 0),  # transparent by default
    )
    STYLE.set_type("background_color", Color)

    def __init__(self, parent, size=None, **options):
        
        # TODO : explain why is there an argument size=None since it looks like it's not used

        self.inherit_style(parent, options)

        if size is not None:
            self.style.modify(width=size[0], height=size[1])
        size = self.get_asked_size()

        class ChildrenList(TypedSet):
            """
            Class for an ordered list of children

            Widgets are sort by overlay, you can access to children sorted by their
            position with children.orderedbypos

            For more efficiency, you can access all the Closable children of a Container
            by doing : Container.children.closable -> WeakTypedList(Closable)
            """

            def __init__(children):
                TypedSet.__init__(children, ItemsClass=Widget)

                children._lists = []
                children._strong_refs = set()
                children._sleeping = WeakTypedList(Widget)

            sleeping = property(lambda children: children._sleeping)

            def _add(children, child):
                """
                This method should only be called by the Widget constructor
                """

                assert child.parent == self
                if child.is_sleeping:
                    if child in children.sleeping:
                        LOGGER.warning("{} already sleeping in {}".format(child, children.sleeping))
                        return
                    children.sleeping.append(child)
                    return

                if child in children:
                    LOGGER.warning("{} already in {}".format(child, children))
                    return
                super().add(child)
                children._strong_refs.add(child)
                for list in children._lists:
                    if list.accept(child):
                        list.add(child)

                if child.is_visible:
                    self._warn_change(child.hitbox)

            def add_list(children, name, list):
                """
                make : children.name = list
                a list must have the following methods :
                    - accept(child) -> bool
                    - add(child) -> None
                    - remove(child) -> None
                Each time the container gets a new child, list.add(child) is called
                Each time a container's child gets killed, list.remove(child) is called
                """
                if False in (
                    hasattr(list, "accept"),
                    hasattr(list, "add"),
                    hasattr(list, "remove"),
                ):
                    raise PermissionError("Wrong list argument : {}".format(list))

                list.name = name
                children._lists.append(list)
                setattr(children, name, list)

            def _remove(children, child):

                if child.is_sleeping:
                    children._sleeping.remove(child)
                else:
                    super().remove(child)
                    for list in children._lists:
                        if list.accept(child):
                            list.remove(child)

                if child.is_visible:
                    self._warn_change(child.hitbox)

                def find(f, key, start):
                    res = start
                    for child2 in children:
                        res = f(child2.__getattribute__(key), res)
                    return res
                """if child.left == 0:
                    dx = find(min, "left", self.w)
                    if children:
                        self.move(dx=dx)
                        for child2 in children:
                            child2.move(dx=-dx)
                    self.resize_width(self.w - dx)
                if child.top == 0:
                    dy = find(min, "top", self.h)
                    if children:
                        self.move(dy=dy)
                        for child2 in children:
                            child2.move(dy=-dy)
                    self.resize_height(self.h - dy)"""
                # if child.right == self.auto_hitbox.w or child.bottom == self.auto_hitbox.h:
                #     self._adapt(self.widgets)

            def remove(children):
                raise PermissionError

            def get_irunning(children):
                for child in children.runables_always:
                    if child.is_running:
                        yield child
                for child in children.runables_at_frame:
                    if child.is_running:
                        yield child
            running = property(lambda children: tuple(children.get_irunning()))
        self._children = ChildrenList()

        # Only layers can guarantie the overlay
        layersmanager_class = LayersManager
        if "layersmanager_class" in options:
            layersmanager_class = options.pop("layersmanager_class")
            assert issubclass(layersmanager_class, LayersManager)
        self.layers_manager = layersmanager_class(self)
        self.layers = self.layers_manager.layers
        self.children.add_list("layers_manager", self.layers_manager)

        self._children_to_paint = WeakTypedSet(Widget)  # a set cannot have two same occurences
        class RectsToUpdate(TypedSet):
            def __init__(set):
                TypedSet.__init__(set, ItemsClass=tuple)
                super().add((0, 0) + size)
            def add(set, rect):
                rect = self.auto_hitbox.clip(rect)
                if rect[2:] == (0, 0): return
                with paint_lock:
                    super().add(tuple(rect))
        self._rects_to_update = None
        # self._rects_to_update = RectsToUpdate()
        self._rect_to_update = None
        self._requests = PackedFunctions()  # using PackedFunctions allow to set an owner for a request

        class PositionSortingList(TypedList):
            def __init__(self, *args, **kwargs):
                TypedList.__init__(self, *args, **kwargs)
            def add(self, p_object):
                super().append(p_object)
                self.sort()
            def insert(self):
                raise PermissionError
            def sort(self):
                super().sort(key=lambda c: (c.abs.top, c.abs.left))
        self.children.add_list("closables", TypedSet(Closable))
        self.children.add_list("containers", TypedSet(Container))
        self.children.add_list("openables", TypedSet(Openable))
        self.children.add_list("focusables", PositionSortingList(Focusable))
        # self.children.add_list("sorted_by_pos", PositionSortingList(Widget))

        # BACKGROUND
        background_color = self.style["background_color"]
        if background_color is None: background_color = (0, 0, 0, 0)
        self._background_color = Color(background_color)
        surf = pygame.Surface(size, pygame.SRCALPHA)

        ResizableWidget.__init__(self, parent, surface=surf, **options)

        if self.is_hidden:
            self.set_dirty(1)

    all_children = property(lambda self: list(self._children) + list(self._children.sleeping),
                            doc="Awake and sleeping children")
    background_color = property(lambda self: self._background_color)
    children = property(lambda self: self._children,
                        doc="Awake children")
    default_layer = property(lambda self: self.layers_manager.default_layer)

    def _add_child(self, child):
        self.children._add(child)

    def _dirty_child(self, child, dirty):
        """
        Should only be called by Widget.send_paint_request()
        """
        try:
            assert (child in self.children) or child is self  # for scenes
        except AssertionError as e:
            raise e
        assert dirty in (0, 1, 2)

        child._dirty = dirty
        if dirty:
            self._children_to_paint.add(child)
        elif child in self._children_to_paint:
            self._children_to_paint.remove(child)

    def _flip(self):
        """Update all the surface"""

        if self.is_hidden:  return
        self._flip_without_update()
        self.parent._warn_change(self.hitbox)

    def _flip_without_update(self):
        """Update all the surface, but don't prevent the parent"""

        if self.is_hidden:  return

        with paint_lock:
            # self._rects_to_update.clear()
            self._rect_to_update = None

            self.surface.fill(self.background_color)

            for layer in self.layers:
                for child in layer.visible:
                    try:
                        self.surface.blit(child.surface, child.hitbox.topleft)
                    except pygame.error as e:
                        # can be raised from a child.surface who is a subsurface from self.surface
                        assert child.surface.get_parent() is self.surface
                        child._flip_without_update()  # overdraw child.hitbox

            if debug_screen_updates:
                LOGGER.info("update in {} :  {}".format(self, self.auto))

    def _find_place_for(self, child):

        if child.layer is None:
            self.layers_manager.set_layer_for(child)
        return child.layer._find_place_for(child)

    def _move(self, dx, dy):

        with paint_lock:
            super()._move(dx, dy)
            for child in tuple(self.children):
                child._update_from_parent_movement()

    def _remove_child(self, child):
        self.children._remove(child)

    def _update_rect(self):
        """
        How to update a given portion of the application ?

        This method is the answer.
        This container will update by himself the portion to update, storing the result
        into its surface. Then, it ask to its parent to update the same portion,
        and its parent will use this container surface.

        But how can a container update its surface ?

        The container create a surface (rect_background) at the rect size. This new surface
        is gonna replace a portion of the container surface corresponding to the rect to
        update, once every child have been blited on it.

        if all is set, will update all of the container's hitbox


        surface :         --------- - - - -----------------------

                                        ^
                                        |

        rect_background :            -------
                                     :     :
                                     :     :
                                     :     :
        child3 :        -------------
                                     :     :
        child2 :                ------------- - - - -             <- The solid line is hitbox, the dotted plus solid line is rect
                                     :     :
        child1 :                ------------------------------
                                     :     :
        background :        --------------------------------------  <- background is filled with background_color
                                     :     :
                                     :     :
                                     :     :
        rect to update :             :-----:

        """

        if self._rect_to_update is None: return
        if self.is_hidden:  return

        with paint_lock:
            # rects_to_update = tuple(self._rects_to_update)
            rect = self._rect_to_update
            self._rect_to_update = None
            # self._rects_to_update.clear()
            # for rect in rects_to_update:

            """
            Si, lorsqu'un un enfant change, un pixel passe de plein a transparent,
            il faut aller chercher la couleur de derriere
            C'est soit background_color, soit un autre enfant

            Si background_color a de la transparence, alors il ne suffit pas de faire un
            blit des enfants sur la surface definie par rect, il faut aussi la
            reinitialiser afin de ne pas superposer la transparence
            """
            self.surface.fill(self.background_color, rect)

            for layer in self.layers:
                for child in layer.visible:
                    if child.hitbox.colliderect(rect):

                        try:
                            # collision is relative to self
                            collision = child.hitbox.clip(rect)
                            self.surface.blit(
                                child.surface.subsurface(
                                    (collision.left - child.rect.left, collision.top - child.rect.top) + collision.size),
                                collision.topleft
                            )
                        except pygame.error as e:
                            # can be raised from a child.surface who is a subsurface from self.surface
                            assert child.surface.get_parent() is self.surface
                            child._flip_without_update()  # overdraw child.hitbox

            if debug_screen_updates:
                LOGGER.info("update in {} :  {}".format(self, rect))

            self._warn_parent(rect)

    def _warn_change(self, rect):
        """Request updates at rects referenced by self"""

        rect = self.auto_hitbox.clip(rect)
        if rect[2:] == (0, 0): return
        if self._rect_to_update is None:
            self._rect_to_update = pygame.Rect(rect)
        else:
            self._rect_to_update.union_ip(rect)
        # self._rects_to_update.add(rect)

    def _warn_parent(self, rect):
        """Request updates at rects referenced by self"""

        self.parent._warn_change(
            (self.rect.left + rect[0], self.rect.top + rect[1]) + tuple(rect[2:])
        )

    def adapt(self, children, padding=0, vertically=True, horizontally=True):
        """
        Resize in order to contain every widget in children
        Only use padding.right and padding.bottom, because it is not supposed to move children
        """

        list = tuple(children)
        padding = MarginType(padding)
        self.resize(
            (max(comp.right for comp in list)+padding.right
             if list else padding.right) if horizontally else self.w,
            (max(comp.bottom for comp in list)+padding.bottom
             if list else padding.bottom) if vertically else self.h
        )

    def asleep_child(self, child):

        if child.is_sleeping:
            return

        assert child in self.children

        child._memory.need_appear = child.is_visible

        # WARNING : trying new order for the temporary layer, so it is not killed if its only widget is set asleep
        # OLD :
        # self.children._remove(child)
        # self.children.sleeping.append(child)

        self.children.sleeping.append(child)
        self.children._remove(child)
        child.hide()
        child._is_sleeping = True
        child.signal.ASLEEP.emit()

    def container_close(self):

        for cont in self.children.containers:
            cont.container_close()
        for child in tuple(self.children.closables):  # tuple prevent from killing closables
            child.close()

    def container_open(self):

        for cont in self.children.containers:
            cont.container_open()
        for child in self.children.openables:
            child.open()

    def container_paint(self):

        for cont in self.children.containers:
            cont.container_paint()

        if self._children_to_paint:
            for child in tuple(self._children_to_paint):
                if child.is_visible:
                    child.paint()
                    if child._dirty == 1:
                        child._dirty = 0
                        self._children_to_paint.remove(child)
                    # LOGGER.debug("Painting {} from container {}".format(child, self))

        if self.dirty == 0:  # else, paint() is called by parent
            self._update_rect()

    def container_exec_requests(self):

        self._requests()
        self._requests.clear()

    def fit(self, layer):

        assert layer in self.layers
        self.resize(max(c.right for c in layer), max(c.bottom for c in layer))

    def has_layer(self, layer_name):
        return layer_name in (layer.name for layer in self.layers)

    def kill(self):

        self.hide()
        for child in tuple(self.all_children):
            child.kill()
        super().kill()

    def paint(self, recursive=False, only_containers=True, with_update=True):

        if recursive:
            for c in self.children:
                if isinstance(c, Container):
                    c.paint(recursive, only_containers, with_update=False)
                elif not only_containers:
                    c.paint()
        if with_update:
            self._flip()
        else:
            self._flip_without_update()

    def resize(self, w, h):

        if self.has_locked.width: w = self.w
        if self.has_locked.height: h = self.h
        if (w, h) == self.size: return

        need_alpha = pygame.SRCALPHA if self.background_color.has_transparency() else 0
        with paint_lock:
            super().set_surface(pygame.Surface((w, h), need_alpha))
            self._flip_without_update()

    def send_request(self, request, owner=None):

        assert callable(request)
        self._requests.add(request, owner=None)

    def set_always_dirty(self):
        """Lock self.dirty to 2, cannot go back"""

        self.set_dirty(2)
        self.set_dirty = lambda dirty: None
        self._warn_change = lambda rect: None
        # WARNING : this function is dirty...

    def set_background_color(self, *args, **kwargs):

        self._background_color = Color(*args, **kwargs)
        self.send_paint_request()

    def set_surface(self, surface):

        raise PermissionError("A Container manage its surface itself (it is the addition of its child surfaces)")

    def wake_child(self, child):

        if child.is_awake:
            return

        assert child in self.children.sleeping

        self.children.sleeping.remove(child)
        child._is_sleeping = False
        self.children._add(child)

        if child._memory.need_start_animation:
            child.start_animation()

        if child._memory.need_appear:
            child.show()

        child.send_paint_request()

        child._memory.need_appear = None
        child._memory.need_start_animation = None

        child.signal.WAKE.emit()





from .textedit import TextEdit, keyboard


class LineEdit(TextEdit):
    """
    LineEdit is a Text who only contains 1 line
    """
    # TODO : a lot of reworks and unit tests
    # TODO : LineEdit.line instead of .lines ?

    # def __init__(self, *args, **kwargs):

    #     TextEdit.__init__(self, *args, **kwargs)

    def handle_keydown(self, key):

        if key != keyboard.RETURN:
            self.cursor.handle_keydown(key)





from baopig.pybao.objectutilities import Object
from baopig.pybao.issomething import *


class RessourcePack:

    def config(self, **kwargs):

        for name, value in kwargs.items():
            self.__setattr__('_'+name, value)


class FontsRessourcePack(RessourcePack):

    def __init__(self,
        file=None,
        height=15,
        color=(0, 0, 0),
    ):

        assert is_color(color)

        self._file = file
        self._height = height
        self._color = color

    file = property(lambda self: self._file)
    color = property(lambda self: self._color)
    height = property(lambda self: self._height)


class ScenesRessourcePack(RessourcePack):

    def __init__(self,
        background_color=(170, 170, 170),
    ):

        assert is_color(background_color)

        self._background_color = background_color

    background_color = property(lambda self: self._background_color)


# TODO : ButtonRessourcePack.style.create_surface(size)

class _RessourcePack:

    def __init__(self):

        self.font = FontsRessourcePack()
        self.scene = ScenesRessourcePack()

ressources = _RessourcePack()





from .widget import Widget


class ResizableWidget(Widget):
    """
    Class for widgets who can be resized

    resize() method must be implemented
    """

    STYLE = Widget.STYLE.substyle()
    STYLE.create(
        width = None,  # must be filled
        height = None,  # must be filled
    )

    def _update_size(self):

        self.resize(*self.get_asked_size())

    def get_asked_size(self):

        size = (self.style["width"], self.style["height"])
        with_percentage = False
        for coord in size:
            if isinstance(coord, str):
                # hard stuff
                assert coord[-1] == '%'
                with_percentage = True
            else:
                assert isinstance(coord, (int, float)), f"Wrong value in size : {coord} (must be a number)"
                assert coord >= 0, f"Wrong value in size : {coord} (must be positive)"

        if with_percentage:
            size = list(size)
            for i, coord in enumerate(size):
                if isinstance(coord, str):
                    size[i] = self.parent.size[i] * int(coord[:-1]) / 100

                    # self.origin.reference.signal.RESIZE.connect(hardcoder)
                    # TODO WARNING : if the origin's reference gets changed, this is not updated

        return size

    def inherit_style(self, theme, options=None, **kwargs):

        size = None
        if options and "size" in options: size = options.pop("size")
        if "size" in kwargs: size = kwargs.pop("size")

        super().inherit_style(theme, options, **kwargs)

        if size is not None:
            self.style._setitem("width", size[0])
            self.style._setitem("height", size[1])

    def resize(self, w, h):
        """
        Here you set up your new component's surface
        """

    def resize_height(self, h):

        if h < 0: h = 0
        if h == self.h: return
        self.style.modify(height=h)
        self.resize(self.w, h)

    def resize_width(self, w):

        if w < 0: w = 0
        if w == self.w: return
        self.style.modify(width=w)
        self.resize(w, self.h)

    h =                 property(lambda self: self._rect.h, resize_height)
    height =            property(lambda self: self._rect.height, resize_height)
    size =              property(lambda self: self._rect.size, resize)
    w =                 property(lambda self: self._rect.w, resize_width)
    width =             property(lambda self: self._rect.width, resize_width)




from baopig import *


class TesterScene(Scene):

    def __init__(self, app, ContentZoneClass):

        # TODO : finish to automatically collect every UT_Scene
        # TODO : button Home

        Scene.__init__(self, app, size=(900, 600), name=str(ContentZoneClass))
        self.sections = []
        # TODO : automatically pack zones with margin

        # Menu
        self.menu_zone = Zone(self, size=(self.w-20, 25), pos=(10, 10), name="menu")
        GridLayer(self.menu_zone, nbrows=1)  # TODO : margin
        Button(self.menu_zone, "< BACK", col=0, command=PrefilledFunction(app.open, "UTMenu_Scene"))
        def try_it_yourself():
            if self.try_it_yourself.is_visible:
                self.try_it_yourself.hide()
            else:
                self.try_it_yourself.show()
        Button(self.menu_zone, "Try it yourself !", col=1,
               command=try_it_yourself)

        # Section
        self.sections_zone = Zone(self, size=(self.w/2-15, self.h-self.menu_zone.h-20), background_color="lightgray",
                                  pos=(10, self.menu_zone.bottom + 10), name="sections")
        # TODO : MenuLayer or ColumnLayer, just a Grid with one column
        GridLayer(self.sections_zone, name="sections_layer", col_width=self.sections_zone.w, nbcols=1)
        self.add_section(title="NO TEST SECTION YET", tests=[])

        # Try it yourself
        self.try_it_yourself = None
        self._init_try_it_yourself()

        # Content
        self.content = ContentZoneClass(self, size=(self.w/2-15, self.h-self.menu_zone.h-20), name="content",
                      pos=(-10, self.menu_zone.bottom + 10), sticky="right")

    def _init_try_it_yourself(self):

        code = """# Default code for Try it yourself

from baopig import *

app = Application()
scene = Scene(app)

def click():
    print("Hello world")
b = Button(scene, "Hello world", command=click)

app.launch()"""

        self.try_it_yourself = Zone(self, size=(self.w/2-15, self.h-20), background_color="lightgray",
                                    pos=(10, self.menu_zone.bottom + 10), name="try_it_yourself")
        self.try_it_yourself.hide()
        self.try_it_yourself.code = TextEdit(self.try_it_yourself, width=self.try_it_yourself.w,
                                                 text=code, font=Font(file="monospace"))  # TODO : font_file

        self.try_it_yourself.console = Text(self.try_it_yourself, pos=(0, "50%"))

        # TODO : TextEdit and LineEdit
        # TODO : CodeEdit
        def run():
            import tempfile
            with tempfile.NamedTemporaryFile(suffix=".py", mode="w") as script:
                script.write(self.try_it_yourself.code.text)
                script.seek(0)

                import subprocess
                with tempfile.TemporaryFile(mode="r") as output_file:
                    subprocess.call("python3 {}".format(script.name), shell=True, stdout=output_file, stderr=output_file)
                    output_file.seek(0)
                    output = output_file.read()

                    # proc = subprocess.Popen(['python3', script.name], stdout=output_file, shell=True)
                    # data = proc.communicate()
                    # # proc.run()
                    # output_file.seek(0)
                    # output = output_file.read()
                    self.try_it_yourself.console.set_text(str(output))

        self.try_it_yourself.run = Button(self.try_it_yourself, "RUN",
                                          sticky="right", command=run, catching_errors=False)

    def add_section(self, title, tests):

        if self.sections and self.sections[0] == ["NO TEST SECTION YET", []]:
            self.sections.pop(0)
            for temp in tuple(self.sections_zone.children):
                temp.kill()

        self.sections.append([title, tests])
        ressources.font.config(file="Gill Sans MT")
        Text(self.sections_zone, "--- SECTION {} : {} ---".format(len(self.sections), title),
             font_height=ressources.font.height+2, bold=True,
             row=len(self.sections_zone.children), max_width=self.sections_zone.w)
        for i, text in enumerate(tests):
            # TODO : CheckBoxes
            Text(self.sections_zone, text="TEST {} : ".format(i+1) + text,  # {:0>2} for 01, 02...
                 row=len(self.sections_zone.children), max_width=self.sections_zone.w)
        Text(self.sections_zone, "", row=len(self.sections_zone.children))

    def set_code(self, code):
        self.try_it_yourself.code.set_text(code)






import pygame
from baopig._lib import *
from .text import Text, DynamicText


class SliderBloc(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        width = -1,   # use length and wideness instead
        height = -1,  # use length and wideness instead
        border_width = 3,
        pos_location = "left",
        pos_ref_location = "left",
    )
    STYLE.create(
        length = 16,
        wideness = 14,
    )

    def __init__(self, slider):

        assert isinstance(slider, Slider)

        self._border_width = 0
        self.inherit_style(slider)  # anticipated inheritance
        self._border_width = self.style["border_width"]
        # TODO : if the slider goes vertically, switch length and wideness in the following line
        self.style.modify(width=self.style["length"], height=self.style["wideness"])

        Rectangle.__init__(self, slider)

        self._max_index = None  # TODO : remove

    slider = property(lambda self: self._parent)
    x_max = property(lambda self: self._max_index)

    def update(self):

        self.x = self.slider.get_pourcent() * self.x_max


class SliderBar(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        pos_location = "center",
        pos_ref_location = "center",
        width = -1,  # TODO : delete, but is it possible ? maybe not
        height = -1,  # TODO : delete
        color = (0, 0, 0, 64),
        border_width = 1,
    )
    STYLE.create(
        length = 150,
        wideness = 10,
    )
    # NOTE : we replaced width/height with length/wideness, so it is easier to code vertical sliders

    def __init__(self, slider):

        assert isinstance(slider, Slider)

        style = slider.get_style_for(self.__class__)
        self.inherit_style(slider)
        # TODO : if the slider goes vertically, switch length and wideness in the following line
        self.style.modify(width=style["length"], height=style["wideness"])

        Rectangle.__init__(self, slider)


class Slider(Container, Linkable, Hoverable):
    """Widget that contains a bar and a slideable bloc"""

    STYLE = Container.STYLE.substyle()
    STYLE.modify(
        width = -1,   # don't use them
        height = -1,  # don't use them
    )
    # NOTE : On peut facilement se tromper en laissant width et height alors qu'on devrait utiliser bar_size
    STYLE.create(
        has_indicator = True,
        bloc_class = SliderBloc,
        bar_class = SliderBar,
        axis = "x",
    )
    STYLE.set_constraint("axis", lambda val: val in ("x", "y"), "must be 'x' or 'y'")

    STYLE.set_type("has_indicator", bool)
    STYLE.set_constraint("bloc_class", lambda val: issubclass(val, SliderBloc))
    STYLE.set_constraint("bar_class", lambda val: issubclass(val, SliderBar))

    def __init__(self, parent, minval, maxval,
                 bloc_class=None, bar_class=None, has_indicator=None, bar_size=None,
                 defaultval=None, step=None, title=None, printed_title=False, **options):

        if "size" in options:
            raise PermissionError("Use bar_size instead of size")

        self.inherit_style(parent, options)

        if defaultval is None: defaultval = minval

        assert minval < maxval, f"There must be a positive difference between minval and maxval " \
                                f"(minval : {minval}, maxval : {maxval})"
        assert minval <= defaultval <= maxval, f"The defaultval must be included between minval and maxval " \
                                             f"(minval : {minval}, maxval : {maxval}, defaultval : {defaultval})"
        if step is not None: assert step > 0

        if bar_size is not None:  # TODO : bar_style
            self.set_style_for(self.style["bar_class"], width=bar_size[0], height=bar_size[1])

        bar_style = self.get_style_for(self.style["bar_class"])
        # TODO : if the slider goes vertically, switch length and wideness in the following line
        self.style.modify(width=bar_style["length"], height=bar_style["wideness"])
        Container.__init__(self, parent, **options)
        Linkable.__init__(self)
        Hoverable.__init__(self)

        self._minval = minval
        self._maxval = maxval
        self._range = self.maxval - self.minval
        self._defaultval = self._val = defaultval
        self._step = step
        self._link_origin = None  # the link x position, relative to self
        self._axis = self.style["axis"]

        self.bar = self.style["bar_class"](self)
        self.bloc = self.style["bloc_class"](self)

        self.resize(
            self.bar.width + max(0, self.bloc.border_width - self.bar.border_width) * 2,
            max(self.bar.height, self.bloc.height),
        )
        self._max_bloc_index = self.bloc._max_index = self.width - self.bloc.width
        self.bloc.update()

        self.create_signal("NEW_VAL")

        if self.style["has_indicator"]:
            get_indicator_text = lambda: self.val
            if title:
                if printed_title:
                    self.title = Text(
                        self, title,
                        color=(96, 96, 96), font_height=int((self.bar.height - self.bar.border_width*2) * .9), bold=True,
                        sticky="center", touchable=False
                    )
                get_indicator_text = lambda: title + f" : {self.val}"
            self.set_indicator(get_text=get_indicator_text)

    axis = property(lambda self: self._axis)
    defaultval = property(lambda self: self._defaultval)
    maxval = property(lambda self: self._maxval)
    minval = property(lambda self: self._minval)
    range = property(lambda self: self._range)
    step = property(lambda self: self._step)
    val = property(lambda self: self._val)

    def _update_val(self, val=None, x=None):

        assert (val is None) != (x is None)
        if x is not None:
            if x == self.bloc.x: return
            # val = x * (max - min) / max_index + min
            val = x * self.range / self._max_bloc_index + self.minval
        if self.step is not None:
            def cut(n, l):
                # print(n, l, float(("{:." + str(l-1) + "e}").format(n)))
                return float(("{:." + str(l-1) + "e}").format(n))
            val = round((val - self.minval) / self.step) * self.step + self.minval
            if isinstance(self.step, float):
                val = cut(val, len(str(self.step % 1)) - 2 + len(str(int(val))))
            if isinstance(val, float) and val.is_integer():  # remove .0 for the beauty of the indicator
                val = int(val)
            if val >= self.maxval:  # not else, because step can make val go to maxval or higher
                val = self.maxval

        if val == self.val: return
        assert self.minval <= val <= self.maxval, f"{val}, {self.maxval}, {self.minval}"

        self._val = val
        # x = (val - min) / (max - min) * max_index
        self.bloc.update()
        if self.bloc.x == 0: self._val = self.minval  # prevent approximations
        self.signal.NEW_VAL.emit(self.val)

    def get_pourcent(self):
        """Return the percentage from min to val in the range min -> max"""
        return (self.val - self.minval) / self.range

    def handle_link(self):

        def clamp(val, min_, max_):
            """Clamp f between min and max"""
            return min(max(val, min_), max_)

        if self.bloc.collidemouse():
            self._link_origin = mouse.get_pos_relative_to(self.bloc)[0]
        else:
            self._link_origin = self.bloc.width / 2
            self._update_val(x=clamp(mouse.get_pos_relative_to(self)[0] - self._link_origin,
                                     0, self._max_bloc_index))

    def handle_link_motion(self, link_motion_event):

        def clamp(val, min_, max_):
            """Clamp f between min and max"""
            return min(max(val, min_), max_)

        x = clamp(
            mouse.get_pos_relative_to(self)[0] - self._link_origin,
            0, self._max_bloc_index
        )
        self._update_val(x=x)

    def resize(self, w, h):

        super().resize(w, h)

        bar_margin = max(0, self.bloc.border_width - self.bar.border_width) * 2
        self.bar.resize_width(self.width - bar_margin)

        # TODO : bloc.length, bloc.wideness

    def reset(self):
        """Set the value to defaultval"""
        if self.val == self.defaultval: return
        self._update_val(self.defaultval)

    def set_defaultval(self, val, reset=True):
        """If reset is True, reset the value to defaultval"""

        if val is self.defaultval: return
        assert self.minval <= val <= self.maxval, f"The value must be included between minval and maxval " \
                                f"(minval : {self.minval}, maxval : {self.maxval}, startval : {val})"

        self._defaultval = val
        if reset: self.reset()






import pygame
from baopig.pybao.issomething import *
from baopig.io import mouse
from .utilities import Color, paint_lock, Size
from .widget import Widget
from .resizable import ResizableWidget
from .layer import Layer


class Rectangle(ResizableWidget):
    """
    A Widget who is just a rectangle filled with one color

    If the border_width parameter is filled, the rectangle will not be filled, only its borders
    The border_width is given in pixels
    The border only goes inside the hitbox, not outside
    """

    STYLE = ResizableWidget.STYLE.substyle()
    STYLE.modify(
        width = 30,
        height = 30,
    )
    STYLE.create(
        color = "theme-color-content",
        border_color = "theme-color-border",
        border_width = 0,
    )
    STYLE.set_type("color", Color)
    STYLE.set_type("border_color", Color)
    STYLE.set_type("border_width", int)
    STYLE.set_constraint("border_width", lambda val: val >= 0, "must be positive")

    def __init__(self, parent, **options):

        self.inherit_style(parent, options)
        Widget.__init__(
            self, parent,
            surface=pygame.Surface(self.get_asked_size(), pygame.SRCALPHA),
            **options
        )

        self._color = self.style["color"]
        self._border_color = self.style["border_color"]
        self._border_width = self.style["border_width"]

        self.paint()

    color = property(lambda self: self._color)
    border_color = property(lambda self: self._border_color)
    border_width = property(lambda self: self._border_width)

    def clip(self, rect):
        """
        Clip the rectangle inside another (wich are both relative to the parent)
        """
        new_hitbox = self.hitbox.clip(rect)
        if new_hitbox == self.hitbox: return
        self.move_at(new_hitbox.topleft)
        self.resize(*new_hitbox.size)

    def paint(self):
        """
        If size is set, this method resizes the Rectangle
        """
        self.surface.fill(self.color)
        if self.border_color is not None:
            pygame.draw.rect(self.surface, self.border_color, (0, 0) + self.size,
                             self.border_width * 2 - 1)

    def resize(self, w, h):

        size = w, h
        surface = pygame.Surface(size, pygame.SRCALPHA)
        self.set_surface(surface)
        self.send_paint_request()

    def set_color(self, color=None):

        if color is None: color = (0, 0, 0, 0)
        self._color = Color(color)
        self.send_paint_request()

    def set_border(self, color, width):

        if not isinstance(color, Color):
            color = Color(color)

        assert is_color(color)
        self._border_color = color

        assert 0 <= width
        self._border_width = width

        self.send_paint_request()


class Border_TBR(ResizableWidget):  # TODO : inherit from Rectangle
    """
    A Border is a non filled rectangle. By default, it fits inside 'size' parameter. If
    'surrounding' is set to True, the size will fit inside the border's interior.
    """

    def __init__(self, parent, color, size, width=1, surrounding=False, **kwargs):

        try:                color = pygame.Color(color)
        except ValueError:  color = pygame.Color(*color)
        assert is_size(size)
        assert 0 <= width
        if surrounding:
            size = (size[0] + width * 2, size[1] + width * 2)

        surface = pygame.Surface(size, pygame.SRCALPHA)
        pygame.draw.rect(surface, color, ((0, 0) + size), width * 2 - 1)

        ResizableWidget.__init__(
            self,
            parent=parent,
            surface=surface,
            **kwargs
        )

        self._color = color
        self._interior = pygame.Rect(self.left + width, self.top + width, self.w - width * 2, self.h - width * 2)
        self._surrounding = surrounding
        self._width = width

    color = property(lambda self: self._color)
    interior = property(lambda self: self._interior)
    surrounding = property(lambda self: self._surrounding)
    width = property(lambda self: self._width)

    def collidemouse(self):

        if not super().collidemouse():
            return False

        return not self.interior.collidepoint(mouse.get_pos_relative_to(self.parent))

        mouse_x, mouse_y = mouse.get_pos_relative_to(self.parent)
        return True in (
            mouse_x < self.rect.left + self.width,
            mouse_x >= self.rect.right - self.width,
            mouse_y < self.rect.top + self.width,
            mouse_y >= self.rect.bottom - self.width,
        )

    def config(self, color=None, width=None, surrounding=None):

        if color is not None:
            assert is_color(color)
            self._color = color

        if width is not None:
            assert 0 <= width
            self._width = width

        if surrounding is not None and bool(surrounding) != self._surrounding:
            self._surrounding = bool(surrounding)

        if self.surrounding:
            self.resize(self.size[0] + self.width * 2, self.size[1] + self.width * 2)
        else:
            self.resize(self.size[0] - self.width * 2, self.size[1] - self.width * 2)

        self._interior = pygame.Rect(self.left + width, self.top + width, self.w - width * 2, self.h - width * 2)

    def resize(self, w, h):

        surface = pygame.Surface((w, h), pygame.SRCALPHA)
        pygame.draw.rect(surface, self.color, (self.auto_hitbox), self.width * 2 - 1)
        self.set_surface(surface)


class Highlighter(Widget):
    """
    A Highlighter is a border filled with one color surrounding a target's hitbox
    If the highlighter can be in the target's layer, it is placed in front of the target
    The border is one pixel inside the hitbox, so targets like scenes can be visually
    highlighted
    """

    def __init__(self, parent, target, color, width, **kwargs):

        # assert target.parent is parent
        if parent.parent == parent: parent = target  # target is a scene
        try:                color = pygame.Color(color)
        except ValueError:  color = pygame.Color(*color)

        size = (target.size[0] + width*2-2, target.size[1] + width*2-2)
        surface = pygame.Surface(size, pygame.SRCALPHA)
        pygame.draw.rect(surface, color, ((0, 0) + size), width * 2 - 1)

        Widget.__init__(
            self,
            parent=parent,
            surface=surface,
            pos=(-width+1, -width+1),
            pos_ref=target,
            **kwargs
        )

        self._color = color
        self._target_ref = target.get_weakref()
        self._width = width

        self.set_nontouchable()
        self.connect("config", self.target.signal.RESIZE)

    color = property(lambda self: self._color)
    target = property(lambda self: self._target_ref())
    width = property(lambda self: self._width)

    def config(self, target=None, color=None, width=None):

        with paint_lock:

            if type(target) is tuple:  # old_size from RESIZE signal
                target = None

            if target is not None:
                self.disconnect(emitter=self.target)
                self._target_ref = target.get_weakref()
                self.connect("config", self.target.signal.RESIZE)
                w = self.width if width is None else width
                self.origin.config(pos=(-w+1, -w+1), reference_comp=self.target, from_hitbox=True)

            if color is not None:
                assert is_color(color)
                self._color = color

            if width is not None:
                assert 0 <= width
                self._width = width

            size = (self.target.hitbox.w + self.width*2-2, self.target.hitbox.h + self.width*2-2)
            if size != self.size:
                surface = pygame.Surface(size, pygame.SRCALPHA)
                pygame.draw.rect(surface, self.color, (0, 0) + size, self.width * 2 - 1)
                self.set_surface(surface)


class Sail(Rectangle):

    STYLE = Rectangle.STYLE.substyle()
    STYLE.modify(
        border_width=0,
    )

    def __init__(self, *args, **kwargs):

        Rectangle.__init__(self, *args, **kwargs)
        self.set_nontouchable()
        if not self.parent.has_layer("nontouchable_layer"):
            Layer(self.parent, name="nontouchable_layer", touchable=False)
        self.swap_layer("nontouchable_layer")


class Polygon(Widget):
    """
    Create a Polygon from vertices
    If offset is set, move all vertices by offset
    """

    def __init__(self, parent, color, vertices, width=0, offset=(0, 0), offset_angle=None, **kwargs):

        assert "pos" not in kwargs, "Use offset instead"

        def plus(p1, p2):
            return p1[0] + p2[0], p1[1] + p2[1]
        def minus(p1, p2):
            return p1[0] - p2[0], p1[1] - p2[1]

        if offset_angle:
            import numpy
            rotation_matrix = numpy.array([[numpy.cos(offset_angle) ,-numpy.sin(offset_angle)],
                                           [numpy.sin(offset_angle) ,numpy.cos(offset_angle)]])
            vertices = tuple(rotation_matrix.dot(v) for v in vertices)

        topleft_corner = min(v[0] for v in vertices), min(v[1] for v in vertices)
        verts2 = tuple(minus(v, topleft_corner) for v in vertices)
        surf = pygame.Surface((
            max(v[0] for v in verts2)+1,
            max(v[1] for v in verts2)+1
        ), pygame.SRCALPHA)
        pygame.draw.polygon(surf, color, verts2, width)

        vertices = tuple(plus(offset, v) for v in vertices)
        pos = (
            min(v[0] for v in vertices),
            min(v[1] for v in vertices),
        )

        Widget.__init__(self, parent, surf, pos, **kwargs)

        self._vertices = vertices

    vertices = property(lambda self: self._vertices)


class Line(Widget):

    def __init__(self, parent, color, point_a, point_b, width=1, offset=(0, 0), offset_angle=None, **kwargs):

        assert "pos" not in kwargs, "Use offset instead"

        def plus(p1, p2):
            return p1[0] + p2[0], p1[1] + p2[1]
        def minus(p1, p2):
            return p1[0] - p2[0], p1[1] - p2[1]

        points = point_a, point_b
        if offset_angle:
            import numpy
            rotation_matrix = numpy.array([[numpy.cos(offset_angle) ,-numpy.sin(offset_angle)],
                                           [numpy.sin(offset_angle) ,numpy.cos(offset_angle)]])
            points = tuple(rotation_matrix.dot(p) for p in points)

        topleft_corner = min(p[0] for p in points), min(p[1] for p in points)
        points2 = tuple(minus(p, topleft_corner) for p in points)
        surf = pygame.Surface((
            max(p[0] for p in points2)+1,
            max(p[1] for p in points2)+1
        ), pygame.SRCALPHA)

        pygame.draw.line(surf, color, points2[0], points2[1], width)
        points = tuple(plus(offset, p) for p in points)
        pos = (
            min(p[0] for p in points),
            min(p[1] for p in points),
        )

        Widget.__init__(self, parent, surf, pos, **kwargs)

        self._points = points

    points = property(lambda self: self._points)


class Circle(Widget):

    def __init__(self, parent, color, center, radius, border_width=0, **kwargs):

        if isinstance(radius, float): radius = int(radius)
        assert "pos" not in kwargs, "Use center instead"
        assert isinstance(radius, int)
        if border_width > 1: raise NotImplemented
        surf = pygame.Surface((radius * 2, radius * 2), pygame.SRCALPHA)
        pygame.draw.circle(surf, color, (radius, radius), radius, border_width)
        Widget.__init__(self, parent, surface=surf, pos=center, pos_location="center", **kwargs)

        self._color = color
        self._radius = radius
        self._border_width = border_width

    color = property(lambda self: self._color)
    radius = property(lambda self: self._radius)
    border_width = property(lambda self: self._border_width)

    def set_radius(self, radius):

        if isinstance(radius, float): radius = int(radius)
        assert isinstance(radius, int)
        assert radius >= 0

        self._radius = radius
        surf = pygame.Surface((radius * 2, radius * 2), pygame.SRCALPHA)
        pygame.draw.circle(surf, self.color, (radius, radius), radius, self.border_width)
        self.set_surface(surf)




import time
from baopig._lib import *
from .text import Text
from .button import Button


class DialogAnswerButton(Button):

    def get_dialog(self):
        dialog = self.parent
        while not isinstance(dialog, Dialog):
            if isinstance(dialog, Scene):
                raise TypeError("A DialogAnswerButton must be inside a Dialog")
            dialog = dialog.parent
        return dialog

    dialog = property(get_dialog)

    def validate(self):

        self.dialog._answer(self.text_widget.text)


class DialogButtonsZone(Zone):
    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        pos=(0, -30),
        pos_location="bottom",
        pos_ref_location="bottom",
        background_color=(0, 0, 0, 60),
    )

    def __init__(self, dialog_frame):

        assert isinstance(dialog_frame, DialogFrame)

        choices = dialog_frame.parent.choices
        Zone.__init__(
            self, dialog_frame,
            size=(dialog_frame.w - 60, 46 * ((len(choices) - 1) / 3 + 1)),
        )
        GridLayer(self, nbrows=int((len(choices) - 1) / 3) + 1, nbcols=min(len(choices), 3),
                  row_height=46, col_width=int(self.w / min(len(choices), 3)))
        for i, choice in enumerate(choices):
            assert isinstance(choice, str), "Other types are not implemented"
            self.dialog.style["answerbutton_class"](self, choice, col=i % 3, row=i // 3, sticky="center")

    def get_dialog(self):
        dialog = self.parent
        while not isinstance(dialog, Dialog):
            if isinstance(dialog, Scene):
                raise TypeError("A DialogAnswerButton must be inside a Dialog")
            dialog = dialog.parent
        return dialog

    dialog = property(get_dialog)


class DialogFrame(Zone):
    STYLE = Zone.STYLE.substyle()
    STYLE.modify(
        pos=("50%", "50%"),
        pos_location="center",
        width=450,
        height=300,
        background_color="theme-color-dialog_background",
    )

    def __init__(self, dialog, style):
        assert isinstance(dialog, Dialog)

        self.inherit_style(dialog, style)
        Zone.__init__(self, dialog)

        self.buttons_zone = dialog.style["buttonszone_class"](self)

        self.title_label = Text(
            self, dialog.title,
            font_height=38,
            pos=("50%", 40), pos_location="center"
        )
        bottom = self.title_label.bottom
        if dialog.description is not None:
            self.description_label = Text(
                self, dialog.description,
                font_height=27, max_width=self.w - 60,
                pos=(30, self.title_label.bottom + 15),
            )
            bottom = self.description_label.bottom

        self.resize_height(max(bottom + 50 + self.buttons_zone.height + 10, self.height))


class Dialog(Scene):
    """
    If one_shot is True, this dialog widget will kill itself after the first answer
    """

    STYLE = Scene.STYLE.substyle()
    STYLE.create(
        title="Dialog",
        description=None,
        choices=("Cancel", "Continue"),
        default_choice_index=0,  # focuses the first answer button
        dialogframe_class=DialogFrame,
        buttonszone_class=DialogButtonsZone,
        answerbutton_class=DialogAnswerButton,
    )
    STYLE.set_type("title", str)
    STYLE.set_type("default_choice_index", int)
    STYLE.set_constraint("dialogframe_class", lambda val: issubclass(val, DialogFrame),
                         "must be a subclass of DialogFrame")
    STYLE.set_constraint("buttonszone_class", lambda val: issubclass(val, DialogButtonsZone),
                         "must be a subclass of DialogButtonsZone")
    STYLE.set_constraint("answerbutton_class", lambda val: issubclass(val, DialogAnswerButton),
                         "must be a subclass of DialogAnswerButton")

    def __init__(self, app, title=None, choices=None, description=None, default_choice_index=0,
                 frame_style=None, one_shot=False, background_image=None):

        self.inherit_style(
            app,
            title=title, choices=choices, description=description,
            default_choice_index=default_choice_index,
        )

        Scene.__init__(self, app)

        self.title = self.style["title"]
        self.choices = self.style["choices"]
        self.description = self.style["description"]
        self.default_choice_index = self.style["default_choice_index"]
        assert self.default_choice_index in range(len(self.choices))
        if background_image:
            self.set_style_for(
                self.style["dialogframe_class"],
                background_image=background_image,
            )
        self.frame = self.style["dialogframe_class"](self, frame_style)

        self.answer = None
        self.create_signal("ANSWERED")
        self.one_shot = one_shot

    def pre_open(self):

        app = self.application
        self.hovered_scene = app.focused_scene
        background = app.display.copy()
        sail = pygame.Surface(app.size, pygame.SRCALPHA)
        sail.fill((0, 0, 0, 100))
        background.blit(sail, (0, 0))
        self.set_background_image(background)

    def open(self):
        self._focus(self.frame.buttons_zone.default_layer[self.default_choice_index])

    def _answer(self, ans):
        """Only called by DialogAnswerButton"""
        self.answer = ans
        self.app.open(self.hovered_scene)  # TODO : self.hovered_scene.open()
        self.signal.ANSWERED.emit(self.answer)
        if self.one_shot:
            self.kill()





class NumEntry(Entry):
    """
    A NumEntry is an Entry who only accepts numbers
    It can accept a defined range of numbers
    """

    def __init__(self, parent, min=None, max=None, accepted_numbers=None, accept_negatives=True, *args, **kwargs):

        Entry.__init__(self, parent, *args, **kwargs)

        self._min = min
        self._max = max
        self._accepted_numbers = accepted_numbers
        self._accept_negatives = bool(accept_negatives)

    def accept(self, text):






import pygame
from baopig.pybao.issomething import *
from baopig.ressources import *
from baopig.font import Font
from baopig.io import keyboard
from baopig._lib import ApplicationExit, Clickable, Layer
from baopig._lib import Widget, Container, Sail, Image, Box, Hoverable
from .label import Label
from .text import TextLabel, Text


class AbstractButton2(Label, Clickable, Hoverable):
    """
    Abstract button

    - background color
    - focus
    - link
    - text
    - hover
    """

    STYLE = Label.STYLE.substyle()
    STYLE.modify(
        background_color = "theme-color-content"
    )
    STYLE.create(
        catching_errors = False,
    )

    def __init__(self, parent, command=None, font=None, size=None, margin=None, name=None,
                 background_color=None, catching_errors=None, hover=None, link=None, focus=None, **kwargs):

        self.inherit_style(parent, background_color=background_color, catching_errors=catching_errors)

        if margin is None: margin = 5
        if command is None: command = lambda: None
        # if name is None: name = "<{}({})>".format(self.__class__.__name__, text)

        assert callable(command), "command must be callable"

        Label.__init__(
            self,
            parent=parent,
            text=text,
            font=font,
            size=size,
            margin=margin,  # TODO : padding instead of margin
            text_location="center",
            selectable=False,
            name=name,
            **kwargs
        )
        Hoverable.__init__(self)
        Clickable.__init__(self, catching_errors=self.style["catching_errors"])

        self.command = command  # non protected field
        self._hover_sail_ref = lambda: None
        self._link_sail_ref = lambda: None
        self._focus_rect_ref = lambda: None
        Layer(self, name="nontouchable_layer", touchable=False)
        self.text_component.swap_layer("nontouchable_layer")

        if hover != -1:
            if hover is None:
                self._hover_sail_ref = Sail(
                    parent=self,
                    color=(0, 0, 0, 63),
                    pos=(0, 0),
                    size=self.size,
                    name=self.name+".hover_sail",
                ).get_weakref()
            else:
                self._hover_sail_ref = Image(
                    self, hover, layer="nontouchable_layer", name=self.name+".hover_sail"
                ).get_weakref()
            self.hover_sail.hide()
            self.hover_sail.connect("show", self.signal.HOVER)
            self.hover_sail.connect("hide", self.signal.UNHOVER)

        if focus != -1:
            if focus is None:
                self._focus_rect_ref = Sail(  # TODO : rethink objects with widgets philosophy (everything's a widget)
                    parent=self,
                    color=(0, 0, 0, 0),
                    pos=(0, 0),  # (half_margin_left, half_margin_top),
                    size=self.size,  # (self.w - half_margin_left - half_margin_right, self.h - half_margin_top - half_margin_bottom),
                    name=self.name + ".focus_rect"
                ).get_weakref()
                self.focus_rect.set_border(color=(0, 0, 0), width=1)  # TODO : Border
            else:
                self._focus_rect_ref = Image(
                    self, focus, layer="nontouchable_layer", name=self.name+".focus_sail"
                ).get_weakref()
            self.focus_rect.hide()
            self.focus_rect.connect("show", self.signal.FOCUS)
            self.focus_rect.connect("hide", self.signal.DEFOCUS)
            self.focus_rect.move_behind(self.text_component)

        if link != -1:
            if link is None:
                self._link_sail_ref = Sail(
                    parent=self,
                    color=(0, 0, 0, 63),
                    pos=(0, 0),
                    size=self.size,
                    name=self.name+".link_sail",
                ).get_weakref()
            else:
                self._link_sail_ref = Image(
                    self, link, layer="nontouchable_layer", name=self.name+".link_sail"
                ).get_weakref()
            self.link_sail.hide()
            self.link_sail.connect("show", self.signal.LINK)
            self.link_sail.connect("hide", self.signal.UNLINK)
            self.link_sail.move_behind(self.text_component)

        self._disable_sail_ref = Sail(  # TODO : same as hover, focus and link
            parent=self,
            color=(255, 255, 255, 128),
            pos=(0, 0),
            size=self.size,
            name=self.name + ".disable_sail"
        ).get_weakref()
        self.disable_sail.hide()

    disable_sail = property(lambda self: self._disable_sail_ref())
    focus_rect = property(lambda self: self._focus_rect_ref())
    hover_sail = property(lambda self: self._hover_sail_ref())
    link_sail = property(lambda self: self._link_sail_ref())

    def enable(self):

        self.disable_sail.hide()
        self.hover_sail.lock_visibility(locked=False)
        if self.is_hovered:
            self.hover_sail.show()

    def disable(self):

        self.disable_sail.show()
        self.hover_sail.hide()
        self.hover_sail.lock_visibility(locked=True)
        self.hover_sail.show()
        assert not self.hover_sail.is_visible

    def handle_keydown(self, key):

        if key is keyboard.RETURN:
            self.link_sail.show()
            self.validate()
        elif key is keyboard.TAB:
            self.focus_next()
        elif key in (keyboard.RIGHT, keyboard.DOWN):
            self.focus_next()
        elif key in (keyboard.LEFT, keyboard.UP):
            self.focus_antecedant()
        # TODO : arrows can swap to next focusables

    def handle_keyup(self, key):

        if key is keyboard.RETURN:
            self.link_sail.hide()

    def validate(self, *args, **kwargs):

        self.command(*args, **kwargs)


class ButtonText(Text):  # TODO : das is das ?

    STYLE = Text.STYLE.substyle()
    STYLE.modify(
        align_mode = "left",
    )

    def __init__(self, button, text, **options):

        assert isinstance(button, AbstractButton)
        assert '\n' not in text
        self.inherit_style(button, **options)
        content_rect = button.content_rect

        if content_rect.height < self.style["font_height"]:
            self.style.modify(font_height=content_rect.height)
            # raise ValueError("This text has a too high font for the text area : "
            #                  f"{self.style['font_height']} (maximum={content_rect.height})")
        Text.__init__(
            self, button,
            text=text,
            sticky="center",
            selectable=False,
            **options
        )
        while self.width > content_rect.width:
            if self.font.height == 2:
                raise ValueError(f"This text is too long for the text area : {text} (area={content_rect}), {self.align_mode}, {self.width}")
            self.font.config(height=self.font.height - 1)  # changing the font will automatically update the text


class AbstractButton(Box, Clickable, Hoverable):
    """
    Abstract button

    - background color
    - focus
    - link
    - text
    - hover
    """

    STYLE = Box.STYLE.substyle()
    STYLE.modify(
        width=100,
        height=35,
        background_color = "theme-color-content",
        padding=10,
    )
    STYLE.create(
        catching_errors = False,
    )

    def __init__(self, parent, command=None, name=None,
                 background_color=None, catching_errors=None, hover=None, link=None, focus=None, **options):

        self.inherit_style(parent, options, background_color=background_color, catching_errors=catching_errors)

        if command is None: command = lambda: None

        assert callable(command), "command must be callable"

        Box.__init__(
            self,
            parent=parent,
            name=name,
            **options
        )
        Hoverable.__init__(self)
        Clickable.__init__(self, catching_errors=self.style["catching_errors"])

        self.connect("press", self.signal.LINK)
        self.connect("unpress", self.signal.UNLINK)  # TODO : only self.signal.THING.connect(function)

        self.command = command  # non protected field

        if self.default_layer is None:
            self.layers_manager.create_temporary_layer()
        self.behind_lines = Layer(self, weight=self.default_layer.weight-1)
        self.above_lines = Layer(self, weight=self.default_layer.weight+1)

        self._hover_sail_ref = lambda: None
        self._link_sail_ref = lambda: None
        self._focus_rect_ref = lambda: None
        # Layer(self, name="nontouchable_layer", touchable=False)
        # self.text_component.swap_layer("nontouchable_layer")

        if hover != -1:
            if hover is None: hover = 63
            if isinstance(hover, int):
                self._hover_sail_ref = Sail(
                    parent=self,
                    color=(0, 0, 0, hover),
                    pos=(0, 0),
                    size=self.size,
                    name=self.name+".hover_sail",
                ).get_weakref()
            else:
                self._hover_sail_ref = Image(
                    self, hover, layer="nontouchable_layer", name=self.name+".hover_sail"
                ).get_weakref()
            self.hover_sail.hide()
            self.hover_sail.connect("show", self.signal.HOVER)
            self.hover_sail.connect("hide", self.signal.UNHOVER)
            self.hover_sail.swap_layer(self.above_lines)

        if focus != -1:
            if focus is None:
                self._focus_rect_ref = Sail(
                    parent=self,
                    color=(0, 0, 0, 0),
                    pos=(0, 0),  # (half_margin_left, half_margin_top),
                    size=self.size,  # (self.w - half_margin_left - half_margin_right, self.h - half_margin_top - half_margin_bottom),
                    border_color="theme-color-border",
                    border_width=1,
                    name=self.name + ".focus_rect"
                ).get_weakref()
                # self.focus_rect.set_border(color=, width=1)  # TODO : Border
            else:
                self._focus_rect_ref = Image(
                    self, focus, layer="nontouchable_layer", name=self.name+".focus_sail"
                ).get_weakref()
            self.focus_rect.hide()
            self.focus_rect.connect("show", self.signal.FOCUS)
            self.focus_rect.connect("hide", self.signal.DEFOCUS)
            self.focus_rect.swap_layer(self.behind_lines)

        if link != -1:
            if link is None:
                self._link_sail_ref = Sail(
                    parent=self,
                    color=(0, 0, 0, 63),
                    pos=(0, 0),
                    size=self.size,
                    name=self.name+".link_sail",
                ).get_weakref()
            else:
                self._link_sail_ref = Image(
                    self, link, layer="nontouchable_layer", name=self.name+".link_sail"
                ).get_weakref()
            self.link_sail.hide()
            self.link_sail.connect("show", self.signal.LINK)
            self.link_sail.connect("hide", self.signal.UNLINK)
            self.link_sail.swap_layer(self.behind_lines)

        self._disable_sail_ref = Sail(  # TODO : same as hover, focus and link
            parent=self,
            color=(255, 255, 255, 128),
            pos=(0, 0),
            size=self.size,
            name=self.name + ".disable_sail"
        ).get_weakref()
        self.disable_sail.hide()
        self.disable_sail.swap_layer(self.above_lines)

        if isinstance(hover, int) and hover != -1:
            hidden = self.is_hidden
            if hidden:
                if self.has_locked.visibility:
                    raise NotImplementedError
                self.show()
            self.paint()  # cannot paint if not visible
            if hidden:
                self.hide()
            self.hover_sail.surface.blit(self.surface, (0, 0), special_flags=pygame.BLEND_RGBA_MIN)

    disable_sail = property(lambda self: self._disable_sail_ref())
    focus_rect = property(lambda self: self._focus_rect_ref())
    hover_sail = property(lambda self: self._hover_sail_ref())
    link_sail = property(lambda self: self._link_sail_ref())

    def enable(self):

        self.disable_sail.hide()
        self.hover_sail.lock_visibility(locked=False)
        if self.is_hovered:
            self.hover_sail.show()

    def disable(self):

        self.disable_sail.show()
        self.hover_sail.hide()
        self.hover_sail.lock_visibility(locked=True)
        self.hover_sail.show()
        assert not self.hover_sail.is_visible

    def handle_keydown(self, key):

        if key is keyboard.RETURN:
            self.link_sail.show()
            self.validate()
        elif key is keyboard.TAB:
            self.focus_next()
        elif key in (keyboard.RIGHT, keyboard.DOWN):
            self.focus_next()
        elif key in (keyboard.LEFT, keyboard.UP):
            self.focus_antecedant()
        # TODO : arrows can swap to next focusable

    def handle_keyup(self, key):

        if key is keyboard.RETURN:
            self.link_sail.hide()

    def press(self):
        """Stuff to do when the button is pressed"""

    def unpress(self):
        """Stuff to do when the button is unpressed"""

    def validate(self, *args, **kwargs):

        self.command(*args, **kwargs)


class Button(AbstractButton):
    """
    Un Button est un bouton classique, avec un text
    Sa couleur d'arriere plan par defaut est (64, 64, 64)
    """

    STYLE = AbstractButton.STYLE.substyle()
    STYLE.create(
        text_class = ButtonText,
        text_style = {},
    )
    STYLE.set_constraint("text_class", lambda val: issubclass(val, ButtonText))

    def __init__(self, parent, text=None,
                 command=None, background_color=None, **kwargs):

        AbstractButton.__init__(
            self,
            parent=parent,
            command=command,
            background_color=background_color,
            **kwargs
        )
        if text is not None:
            assert isinstance(text, str)
            self.text_widget = self.style["text_class"](self, text=text, **self.style["text_style"])
            if self.name == "NoName": self._name = text

    def copy(self):

        return Button(
            parent=self.parent,
            text=self.text,
            pos=self.pos,
            w=self.w,
            h=self.h,
            margin=0,  # TODO : self.margin
            font_size=self.font_size,
            font_color=self.font_color,
            command=self.command,
            background_color=self.background.color,
            name="Copied"+self.name
        )






import sys
import os
import time
import threading
import pygame
from baopig.pybao.objectutilities import PrefilledFunction, History, TypedList, Object
from baopig._debug import infinite_fps, debug_global_fps
from baopig.time.timer import RepeatingTimer
from baopig._lib import paint_lock
from .thread import ExtraThread, LOGGER


class DrawingThread(ExtraThread):  # TODO : rename as _PainterThread

    def __init__(self, app):

        ExtraThread.__init__(self, app)

        basename = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        self.out_directory = os.path.abspath("out") + os.path.sep + basename + os.path.sep
        self._is_recording = False
        self.record_index = 1

        self._required_fps = None

        # TODO : debug_global_fps (it is a little consuming)
        if debug_global_fps:
            def _tick_fps():
                # being a deque, it manages its data itself
                self.fps_history.append(self.scenes_rendered_during_current_second)
                self.scenes_rendered_during_current_second = 0
            self.scenes_rendered_during_current_second = 0
            self.fps_history = History(int, seq=[], maxlen=500)
            self.fps_history_updater = RepeatingTimer(1, _tick_fps)

        # The clock is the object who freeze enough the thread in order to have maximum app.required_fps FPS
        self.clock = pygame.time.Clock()
        self._can_draw = threading.Event()

    def __del__(self):

        if debug_global_fps:
            self.fps_history_updater.cancel()

    executable_requests = property(lambda self: self._executable_requests)
    is_recording = property(lambda self: self._is_recording)
    required_fps = property(lambda self: self._required_fps)

    def _paint(self):

        # Updating then drawing
        with paint_lock:
            try:
                self.app.focused_scene.container_paint()
            except Exception as e:
                LOGGER.exception(e)

        # FPS Tracer
        """if self.fps_label.is_visible:
            for i, required_fps in enumerate(self.fps_history):
                color = 127
                pygame.draw.line(self.display,
                                 (color, color, color),
                                 (i + 10, self.rect.bottom - 10),
                                 (i + 10, self.rect.bottom - 10 - required_fps * 10))"""

    def get_current_fps(self):
        """

        :return:
        """
        if debug_global_fps:
            if len(self.fps_history) > 0:
                return self.fps_history[-1]
            else:
                return None
        else:
            return self.clock.get_fps()

    def init(self):

        if debug_global_fps:
            self.fps_history_updater.start()

    def screenshot(self):

        os.makedirs(self.out_directory, exist_ok=True)
        screenshot = self.app.display.copy()
        name = time.strftime("%Y.%m.%d-%Hh%M-%S.png", time.localtime())
        print(self.out_directory + name)
        pygame.image.save(screenshot, self.out_directory + name)

    def send_request(self, request):
        """
        Method for objects who need that an action is operated by the main thread

        request is a callable object
        """
        # TODO : remove
        assert callable(request), "A callable is required, got {} instead".format(request)

        self.executable_requests.append(request)

    def set_fps(self, fps):

        assert isinstance(fps, int) and fps > 0
        self._required_fps = fps

    def stop(self):

        super().stop()
        if debug_global_fps:
            self.fps_history_updater.cancel()

    def start_recording(self, only_at_change=False):

        os.makedirs(self.out_directory, exist_ok=True)
        self._is_recording = Object(only_at_change=only_at_change)
        LOGGER.info("Start recording" + (" (only at display updates)" if only_at_change else ""))

    def stop_recording(self):

        self._is_recording = False
        LOGGER.info("Stop recording")

    def update(self):

        self._can_draw.wait()

        try:

            # Threads communication
            # for executable_request in self.executable_requests:
            #     executable_request()
            # self.executable_requests.clear()

            # Drawings
            # TODO : render the display only if at least one run in Application is done
            self._paint()

            # launch time
            if self.app.launch_time is not None:
                import time
                LOGGER.info("{} launched in {} seconds".format(self.app.name, time.time() - self.app.launch_time))
                self.app.launch_time = None

            # record
            if self.is_recording:
                if not self.is_recording.only_at_change:
                    pygame.image.save(self.app.display, self.out_directory + "record_{:0>3}.png".format(self.record_index))
                    self.record_index += 1

            # FPS
            if debug_global_fps:
                self.scenes_rendered_during_current_second += 1
            # NOTE : Pour mieux tester les FPS, on ne fait pas ticker l'horloge
            if infinite_fps is False:
                self.clock.tick(self.required_fps)  # keep the game running slower than the given FPS
            else:
                self.clock.tick(0)
            if self.app._debug_fps:
                pygame.display.set_caption(f"fps : {int(self.clock.get_fps()*100)/100}")
            self._can_draw.clear()

        except pygame.error as e:
            # print(e.__str__())
            if e.__str__() == "video system not initialized":
                self.stop()
                LOGGER.error("The window have been closed")
            else:
                raise e




from .logging import LOGGER


class _ClipBoard:
    def __init__(self):
        self._data = {}

    data = property(lambda self: self._data)

    def get(self, type):
        try:
            return self.data[type]
        except KeyError:
            return None

    def put(self, data):
        self.data[type(data)] = data
        # LOGGER.info("Copied {} object : '{}' inside clipboard".format(type(data).__name__, data))

    def contains(self, type):
        return type in self.data

    def get_types(self):
        return self.data.keys()


clipboard = _ClipBoard()




from baopig import *
import pygame


# TODO : solve : selection error, there seems to be a problem with abs_hitbox or selection_rect
class UT_Text_Scene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app)

        hello = Text(
            parent=self,
            pos=(10, 10),
            text="Hello world"
                 "\nIamaverylongword,canyoureadmecorrectly?"
                 "\nWhat do you want to do ?",
            width=250,
            # mode=Text.FILLED_MODE,  # TODO
        )
        ressources.font.config(file="Arial Narrow Bold Italic.ttf")
        hello2 = Text(
            parent=self,
            pos=(hello.right + 10, 10),
            text="Hello world"
                 "\nIamaverylongword,canyoureadmecorrectly?"
                 "\nWhat do you want to do ?",
            width=hello.width,
            # mode=Text.FILLED_MODE,
        )

        # TODO : hide selection_rect on double clic
        ressources.font.config(file="monospace")
        ressources.font.config(height=60)
        ressources.font.config(color=(0, 100, 0))
        text = Text(
            self,
            pos=hello.bottomleft,
            text="- Bonjour à tous\n"
            "- Bonjour monsieur. Comment allez-vous ?\n"
            "- Très bien merci. Nous allons commencer.\n"
            "- Monsieur, j'avais oublié, j'ai un rendez-vous !\n"
            "- Eh bien filez, ça ira pour cette fois.\n"
            "- Merci monsieur !\n"
            "Et il partit. (vert fonce)",
            width=370,
            font=Font(height=10, file="Arial Narrow Bold Italic.ttf")
        )
        text.font.config(height=25)
        text.set_background_color((255, 255, 255, 128))
        edit = TextEdit(self, pos=text.bottomleft, font=text.font, width=text.max_width, text="Green")
        text.font.config(color=(10, 50, 30))

        ressources.font.config(height=20)
        ressources.font.config(color=(0, 0, 0))
        z = Zone(self, pos=edit.bottomleft, size=("90%", "100%"))
        GridLayer(z)
        for i, file in enumerate((
            "I am written in the default font",
            "monospace",
            "Georgia.ttf",
            "PlantagenetCherokee.ttf",
            "Trebuchet MS Italic.ttf",
            "Times New Roman.ttf",
            "Arial.ttf",
            "Osaka.ttf",
            "Silom",
            "Wingdings 2.ttf",
            "Calibri.ttf",
            "Candara.ttf",
            "Consolas",
            "Verdana",
            "Keyboard.ttf",
            "Gill Sans MT",
            "Palatino Linotype.ttf",
            "JetBrainsMono-Regular.ttf",
            "JetBrainsMono-Medium.ttf",
            "JetBrainsMono-Bold.ttf",
            "JetBrainsMono-ExtraBold.ttf",
            "Nimportequoi.ttf",
        )):
            Text(z, text=file[:-4] if file.endswith(".ttf") else file, font=Font(file=file), row=i)
        #z.default_layer.pack()


class UT_TextEdit_Scene(Scene):

    def __init__(self, app):

        Scene.__init__(self, app, size=(1000, 700))

        self.zone1 = Zone(self, pos=(10, 10), size=(self.w / 2 - 15, self.h - 20), background_color=(100, 100, 100, 50))
        # print(self.zone1)
        self.zone2 = Zone(self, pos=(self.zone1.right + 10, 10), size=(self.w / 2 - 15, self.h - 20), background_color=(100, 100, 100, 50))

        self.zone1.mirrored = TextEdit(self.zone1, text="0123456789012345", width=40, pos=(10, 10))
        self.zone1.d = DynamicText(self.zone1, self.zone1.mirrored.get_text, pos=self.zone1.mirrored.topright)

        text = TextEdit(self.zone2, pos=(10, 10), width=self.zone2.w - 20)

        self.b = Button(self.zone2, text="RUN", pos=text.topright, command=lambda: exec(text.text))
        self.b.origin.config(location="topright")
        text.signal.RESIZE.connect(lambda: self.b.origin.config(pos=text.topright), owner=self.b)
        text.signal.MOTION.connect(lambda: self.b.origin.config(pos=text.topright), owner=self.b)


        scrap = \
            "f = application.focused_scene\n" \
            "c = f.children\n" \
            "b = Button(application.focused_scene, 'HELLO WORLD', command=lambda: print('Hello world'))\n" \
            "Dragable.set_dragable(b)\n" \
            "b2 = b.copy()\n" \
            "Dragable.set_dragable(b2)\n" \
            ""[:-1]
        scrap = \
            "fr = application.focused_scene\n" \
            "et = fr.zone2.default_layer[0]\n" \
            "font = et.font\n" \
            "for i in range(2, 30):\n" \
            "   font.config(height=i)\n" \
            ""[:-1]
        clipboard.put(scrap)


if __name__ == "__main__":
    app = Application(size=(600, 900))
    UT_Text_Scene(app)
    UT_TextEdit_Scene(app)
    app.launch()




from baopig.pybao.objectutilities import PrefilledFunction, Object, TypedDeque
from baopig._lib import *
from .text import Text, _LineSelection


# TODO : LineEntry (with select_all_on_focus and exec_on_defocus)
# TODO : solve arrows
# TODO : presentation text when nothing is in the text ?

class TextEdit(Text, Selector):

    STYLE = Text.STYLE.substyle()
    STYLE.modify(
        width = 100,
        background_color = "theme-color-font_opposite",
    )

    def __init__(self,
        parent,
        text=None,
        pos=None,
        font=None,
        font_color=None,
        font_height=None,
        width=None,
        **kwargs
    ):

        if text is None: text = ""
        if "max_width" in kwargs:
            raise PermissionError("Use width instead of max_width")

        self.inherit_style(parent, width=width)

        Text.__init__(self,
            parent=parent,
            text=text,
            pos=pos,
            font=font,
            color=font_color,
            font_height=font_height,
            max_width=self.style["width"],
            selectable=True,
            **kwargs
        )
        Selector.__init__(self)
        self.enable_selecting(True)

        self._cursor_ref = lambda: None

        self.set_background_color((255, 255, 255))
        self.cursors_layer = Layer(self, Cursor, name="cursors_layer", touchable=False)
        # self.layers.add_layer("cursors", CompsClass=Cursor, touchable=False)
        self.set_selectionrect_visibility(False)

    cursor = property(lambda self: self._cursor_ref())

    def accept(self, text):
        if text == '': return False
        return True

    def cut(self):

        self.copy()
        self.del_selected_data()

    def del_selected_data(self):

        if not self.is_selecting: return
        cursor_index = self.find_index(char_index=self.line_selections[0].index_start,
                                       line_index=self.line_selections[0].line.line_index)
        assert self.is_selecting
        selected_comps = tuple(self.selectables.selected)
        if selected_comps:
            if self.cursor is not None:
                self.cursor.save()
            for line in selected_comps:
                line._text = line.text[:line.selection.index_start] + line.text[line.selection.index_end:]
                line._end = '' if line.selection._is_selecting_line_end else line.end
            line.config()
        self.close_selection()
        self.cursor.config(text_index=cursor_index)

    def end_selection(self, abs_pos, visible=None):

        if self.selection_rect.end != abs_pos:
            super().end_selection(abs_pos, visible)
            # self.selection_rect.end is one pixel lower and on x axis than abs_pos, so we don't use it in this calculation
            pos = (self.selection_rect.end[0] - self.abs.left, self.selection_rect.end[1] - self.abs.top)
            line_index, char_index = self.find_indexes(pos=pos)
            if line_index != self.cursor.line_index or char_index != self.cursor.char_index:
                self.cursor.config(line_index=line_index, char_index=char_index, selecting="done")

    def handle_focus(self):
        assert self.cursor is None

        mouse_pos = mouse.get_pos_relative_to(self.parent)

        w = abs(self.rect.left - mouse_pos[0])
        h = abs(self.rect.top - mouse_pos[1])

        line_index = len(self.lines) - 1
        char_index = len(self.lines[-1].text)

        Cursor(self, line_index=line_index, char_index=char_index)

    def handle_keydown(self, key):

        self.cursor.handle_keydown(key)

    def handle_link(self):
        if not mouse.has_double_clicked:  # else, the cursor follow the selection
            self.cursor.config(text_index=self.find_mouse_index())

    def paste(self, data):

        self.cursor.write(data)

    def set_text(self, text):

        self._selection_start = None
        super().set_text(text)


class HaveHistory:

    def __init__(self):

        """
        An History element is created when :
            - A new text insert
            - A part of text pop
            - Just before a selected data is delete

        An History element store these data :
            - the entire text of parent
            - the cusror indexes (line and char)
            - the selection start and end, if the parent was selecting
        """
        max_item_stored = 50
        self.history = TypedDeque(Object, maxlen=max_item_stored)
        self.back_history = TypedDeque(Object, maxlen=max_item_stored)

    def redo(self):
        """
        Restaure la derniere modification
        """
        if self.back_history:

            backup = self.back_history.pop()  # last element of self.back_history, the current state
            self.history.append(backup)

            backup = self.history[-1]
            self.parent.set_text(backup.text)
            self.config(line_index=backup.cursor_line_index, char_index=backup.cursor_char_index, save=False)
            if backup.selection_start is not None:
                if self.parent.is_selecting:
                    self.parent.close_selection()
                self.parent.start_selection(backup.selection_start)
                self.parent.end_selection(backup.selection_end)

        # else:
        #     LOGGER.info("Cannot redo last operation because the operations history is empty")

    def save(self):

        # if self.parent.is_selecting:
        current = Object(
            text=self.parent.text,
            cursor_line_index=self.line_index,
            cursor_char_index=self.char_index,
            selection_start=self.parent.selection_rect.start if self.parent.selection_rect else None,
            selection_end=self.parent.selection_rect.end if self.parent.selection_rect else None
        )
        self.history.append(current)
        self.back_history.clear()

    def undo(self):
        """
        Annule la derniere modification
        """
        if len(self.history) > 1:  # need at least 2 elements in history

            backup = self.history.pop()  # last element of self.history, which is the state before undo()
            self.back_history.append(backup)

            previous = self.history[-1]
            self.parent.set_text(previous.text)
            self.config(line_index=previous.cursor_line_index, char_index=previous.cursor_char_index, save=False)
            if previous.selection_start is not None:
                if self.parent.is_selecting:
                    self.parent.close_selection()
                self.parent.start_selection(previous.selection_start)
                self.parent.end_selection(previous.selection_end)
        # else:
        #     LOGGER.info("Cannot undo last operation because the operations history is empty")


class Cursor(Rectangle, HaveHistory, RepetivelyAnimated):
    """
    By default, at creation, a cursor is set at mouse position
    """

    def __init__(self, parent, line_index, char_index):

        assert isinstance(parent, TextEdit)
        assert parent.cursor is None

        h = parent.font.height

        Rectangle.__init__(
            self,
            parent=parent,
            pos=(parent.lines[line_index].find_pixel(char_index), parent.lines[line_index].top),
            size=(int(h / 10), h),
            color=ressources.font.color,
            name=parent.name + " -> cursor"
        )
        HaveHistory.__init__(self)
        RepetivelyAnimated.__init__(self, interval=.5)

        self._char_index = None  # index of cursor position, see _Line._chars_pos for more explanations
        self.__line_index = None  # index of cursor line, see Text._lines_pos for more explanations
        self._line = None
        self._text_index = None  # index of cusor in Text.text

        self.parent._cursor_ref = self.get_weakref()
        self.connect("kill", self.parent.signal.DEFOCUS)
        self.swap_layer("cursors_layer")
        self.set_nontouchable()
        self.start_animation()

        self.config(line_index=line_index, char_index=char_index)

    char_index = property(lambda self: self._char_index)
    def _set_line_index(self, li):
        self.__line_index = li
        self._line = self._parent.lines[li]
    _line_index = property(lambda self: self.__line_index, _set_line_index)
    line_index = property(lambda self: self.__line_index)
    line = property(lambda self: self._line)
    text_index = property(lambda self: self._text_index)

    def config(self, text_index=None, line_index=None, char_index=None, selecting=False, save=True):
        """
        Place the cursor at line n° line_index and before the character n° char_index, count from 0
        If text_index is given instead of line_index and char_index, we use parent.find_indexes

        If char_index is at the end of a cutted line (a line too big for the text max_width), then
        the cursor can either be on the end of the line or at the start of the next line, it is
        algorithmically the same. So the object who config the cursor will decide where to place the
        cursor. It can give a float value for text_index (like 5.4) wich mean "Hey, if the cursor is
        at the end of a cutted line, let it move the start of the next one." In this exemple, the
        text_index value will be 5. This works also with char_index = 5.4
        """

        if text_index is not None:
            assert line_index is None
            assert char_index is None
            line_index, char_index = self.parent.find_indexes(text_index=text_index)
            assert text_index == self.parent.find_index(line_index, char_index)
        else:
            assert char_index is not None
            if line_index is None:
                text_index, line_index, char_index = self.parent.find_indexes(line_index=self.line_index, char_index=char_index)
            else:
                text_index = self.parent.find_index(line_index, char_index)

        assert text_index == self.parent.find_index(line_index, char_index)

        if selecting is True and self.parent.selection_rect.start is None:
            pos = self.parent.find_pos(self.text_index)
            abs_pos = self.parent.abs_left + pos[0], self.parent.abs_top + pos[1]
            self.parent.start_selection(abs_pos)

        def fit(v, min, max):
            if v < min: v = min
            elif v > max: v = max
            return v

        self._text_index = fit(text_index, 0, len(self.parent.text))
        self._line_index = fit(line_index, 0, len(self.parent.lines))
        self._char_index = fit(char_index, 0, len(self.line.text))

        # print("Set cursor at :", self.text_index, self.line_index, self.char_index, selecting, save)

        if self.char_index == len(self.line.text_with_end):
            LOGGER.warning("Tricky cursor position")

        if self.get_weakref()._comp is None:
            LOGGER.warning('This component should be dead :', self)

        old_pos = self.topleft
        self.y = self.line.y
        self.x = self.line.find_pixel(self.char_index)

        """if self.x > self.parent.w - self.w:
            dx = self.x - (self.parent.w - self.w)
            self.x -= dx
            self.line.x -= dx  # TODO : scroll text

        if self.x < 0:
            dx = - self.x
            self.x += dx
            self.line.x += dx"""  # TODO : scroll text

        self.start_animation()
        self.show()

        if selecting is "done":
            pass
        elif selecting is True:
            if self.parent.selection_rect.end is None or old_pos != self.pos:
                self.parent.end_selection((self.abs_left, self.abs_top))
        elif selecting is False:
            if self.parent.is_selecting:
                self.parent.close_selection()
        else:
            raise PermissionError

        if save and (not self.history or self.parent.text != self.history[-1].text):
            self.save()

    def handle_keydown(self, key):
        """
        N'accepte que les evenements du clavier
        Si la touche est speciale, effectue sa propre fonction
        Modifie le placement du curseur
        """

        # Cmd + ...
        if keyboard.mod.cmd:
            # Maj + Cmd + ...
            if keyboard.mod.maj:
                if key == keyboard.z:
                    self.redo()
                return
            elif keyboard.mod.ctrl or keyboard.mod.alt:
                return
            elif key == keyboard.d:
            # Duplicate
                selected_data = self.parent.get_selected_data()
                if selected_data == '':
                    selected_data = self.line.real_text
                    self.line.insert(0, selected_data)
                else:
                    self.parent.close_selection()
                    self.line.insert(self.char_index, selected_data)
                self.config(text_index=self.text_index + len(selected_data))
            elif key == keyboard.r:
            # Execute
                try:
                    exec(self.parent.text)
                except Exception as e:
                    LOGGER.warning("CommandError: "+str(e))
            elif key == keyboard.z:
                self.undo()
            elif key in (keyboard.LEFT, keyboard.HOME):
                self.config(self.parent.find_index(line_index=self.line_index, char_index=0), selecting=keyboard.mod.maj)
            elif key in (keyboard.RIGHT, keyboard.END):
                self.config(self.parent.find_index(line_index=self.line_index, char_index=len(self.line.text)), selecting=keyboard.mod.maj)
            elif key == keyboard.UP:
                if self.line_index > 0:
                    self.config(line_index=0,
                                char_index=self.parent.lines[0].find_index(self.rect.left),
                                selecting=keyboard.mod.maj)
            elif key == keyboard.DOWN:
                if self.line_index < len(self.parent.lines)-1:
                    self.config(line_index=len(self.parent.lines)-1,
                                char_index=self.parent.lines[len(self.parent.lines)-1].find_index(self.rect.left),
                                selecting=keyboard.mod.maj)
            return

        # Cursor movement
        if 272 < key < 282 and key != 277:

            if key in (keyboard.LEFT, keyboard.RIGHT):

                if keyboard.mod.alt:  # go to word side
                    if key == keyboard.LEFT:
                        if self.char_index == 0: return
                        self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                        while self.char_index > 0 and \
                                (self.line.text[self.char_index-1] != ' ' or self.line.text[self.char_index] == ' '):
                            self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                    elif key == keyboard.RIGHT:
                        if self.char_index == len(self.line.text): return
                        self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)
                        while self.char_index < len(self.line.text) and \
                                (self.line.text[self.char_index-1] != ' ' or self.line.text[self.char_index] == ' '):
                            self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)
                elif (not keyboard.mod.maj) and self.parent.is_selecting:
                    if key == keyboard.LEFT:
                        self.config(line_index=self.parent.line_selections[0].line_index,
                                    char_index=self.parent.line_selections[0].index_start)
                    elif key == keyboard.RIGHT:
                        self.config(line_index=self.parent.line_selections[-1].line_index,
                                    char_index=self.parent.line_selections[-1].index_end)
                elif key == keyboard.LEFT:
                    self.config(char_index=self.char_index-1,
                                selecting=keyboard.mod.maj)
                    # self.config(text_index=self.text_index - 1, selecting=keyboard.mod.maj)
                elif key == keyboard.RIGHT:
                    self.config(char_index=self.char_index+1,
                                selecting=keyboard.mod.maj)
                    # self.config(text_index=self.text_index + 1, selecting=keyboard.mod.maj)

            elif key in (keyboard.HOME, keyboard.END):
                if key == keyboard.HOME:  # Fn + K_LEFT
                    self.config(self.parent.find_index(line_index=self.line_index, char_index=0), selecting=keyboard.mod.maj)
                elif key == keyboard.END:  # Fn + K_RIGHT
                    self.config(self.parent.find_index(line_index=self.line_index, char_index=len(self.line.text)), selecting=keyboard.mod.maj)

            elif key in (keyboard.UP, keyboard.DOWN):
                if key == keyboard.UP:
                    if self.line_index > 0:
                        self.config(line_index=self.line_index-1,
                                    char_index=self.parent.lines[self.line_index-1].find_index(self.rect.left),
                                    selecting=keyboard.mod.maj)
                if key == keyboard.DOWN:
                    if self.line_index < len(self.parent.lines)-1:
                        self.config(line_index=self.line_index+1,
                                    char_index=self.parent.lines[self.line_index+1].find_index(self.rect.left),
                                    selecting=keyboard.mod.maj)

            elif key in (keyboard.PAGEUP, keyboard.PAGEDOWN):
                if key == keyboard.PAGEUP:
                    if self.line_index > 0:
                        self.config(line_index=0,
                                    char_index=self.parent.lines[0].find_index(self.rect.left),
                                    selecting=keyboard.mod.maj)
                if key == keyboard.PAGEDOWN:
                    if self.line_index < len(self.parent.lines)-1:
                        self.config(line_index=len(self.parent.lines)-1,
                                    char_index=self.parent.lines[len(self.parent.lines)-1].find_index(self.rect.left),
                                    selecting=keyboard.mod.maj)

        # Suppression
        elif key == keyboard.BACKSPACE:
            if self.parent.is_selecting:
                self.parent.del_selected_data()
            elif self.line_index > 0 or self.char_index > 0:
                if self.char_index > 0:
                    self.line.pop(self.char_index-1)
                else:
                    self.parent.lines[self.line_index - 1].pop(-1)
                old = self.text_index
                self.config(text_index=self.text_index - 1)
                assert self.text_index == old - 1

        elif key == keyboard.DELETE:
            if self.parent.is_selecting:
                self.parent.del_selected_data()
            if self.line.end == '\v' and self.char_index == len(self.line.text):
                if self.line_index < len(self.parent.lines) - 1:
                    self.parent.lines[self.line_index + 1].pop(0)
            else:
                self.line.pop(self.char_index)
            self.config(line_index=self.line_index,  # We don't use text_index because, if self.char_index is 0,
                        char_index=self.char_index)  # we want to stay at 0, text_index might send the cursor at the
                                                     # end of the previous line if it is a cutted line

        elif key == keyboard.ESCAPE:
            self.parent.defocus()

        elif keyboard.F1 <= key <= keyboard.F15:
            return

        # Write
        else:
            assert keyboard.last_event.key == key
            unicode = keyboard.last_event.unicode
            if key == keyboard.RETURN:
                unicode = '\n'
            elif key == keyboard.TAB:
                unicode = '    '
            self.write(unicode)

    def write(self, string):

            # Lettres (minuscules et majuscules)
            text = self.parent.text[:self.char_index] + string + self.parent.text[self.char_index:]
            if self.parent.accept(text):

                if self.parent.is_selecting:
                    self.parent.del_selected_data()

                self.line.insert(self.char_index, string)
                self.config(text_index=self.text_index + len(string))

